from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, IncludeCustomEndpoint, \
    UuidEndpoint, ListEndpoint, ChannelIncludeEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.objects_v2.memberships import PNGetMembershipsResult
from pubnub.models.consumer.objects_v2.page import PNPage
from pubnub.structures import Envelope


class PNGetMembershipsResultEnvelope(Envelope):
    result: PNGetMembershipsResult
    status: PNStatus


class GetMemberships(ObjectsEndpoint, UuidEndpoint, ListEndpoint, IncludeCustomEndpoint,
                     ChannelIncludeEndpoint):
    GET_MEMBERSHIPS_PATH = "/v2/objects/%s/uuids/%s/channels"

    def __init__(self, pubnub, uuid: str = None, include_custom: bool = False, limit: int = None, filter: str = None,
                 include_total_count: bool = None, sort_keys: list = None, page: PNPage = None):
        ObjectsEndpoint.__init__(self, pubnub)
        UuidEndpoint.__init__(self, uuid=uuid)
        ListEndpoint.__init__(self, limit=limit, filter=filter, include_total_count=include_total_count,
                              sort_keys=sort_keys, page=page)
        IncludeCustomEndpoint.__init__(self, include_custom=include_custom)
        ChannelIncludeEndpoint.__init__(self)

    def build_path(self):
        return GetMemberships.GET_MEMBERSHIPS_PATH % (self.pubnub.config.subscribe_key, self._effective_uuid())

    def validate_specific_params(self):
        self._validate_uuid()

    def create_response(self, envelope) -> PNGetMembershipsResult:
        return PNGetMembershipsResult(envelope)

    def sync(self) -> PNGetMembershipsResultEnvelope:
        return PNGetMembershipsResultEnvelope(super().sync())

    def operation_type(self):
        return PNOperationType.PNGetMembershipsOperation

    def name(self):
        return "Get Memberships"

    def http_method(self):
        return HttpMethod.GET



from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, ListEndpoint, \
    IncludeCustomEndpoint, IncludeStatusTypeEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.objects_v2.channel import PNGetAllChannelMetadataResult
from pubnub.models.consumer.objects_v2.page import PNPage
from pubnub.structures import Envelope


class PNGetAllChannelMetadataResultEnvelope(Envelope):
    result: PNGetAllChannelMetadataResult
    status: PNStatus


class GetAllChannels(ObjectsEndpoint, ListEndpoint, IncludeCustomEndpoint, IncludeStatusTypeEndpoint):
    GET_ALL_CHANNELS_PATH = "/v2/objects/%s/channels"

    def __init__(self, pubnub, include_custom=False, include_status=True, include_type=True, limit: int = None,
                 filter: str = None, include_total_count: bool = None, sort_keys: list = None, page: PNPage = None):
        ObjectsEndpoint.__init__(self, pubnub)
        ListEndpoint.__init__(self, limit=limit, filter=filter, include_total_count=include_total_count,
                              sort_keys=sort_keys, page=page)
        IncludeCustomEndpoint.__init__(self, include_custom=include_custom)
        IncludeStatusTypeEndpoint.__init__(self, include_status=include_status, include_type=include_type)

    def build_path(self):
        return GetAllChannels.GET_ALL_CHANNELS_PATH % self.pubnub.config.subscribe_key

    def create_response(self, envelope) -> PNGetAllChannelMetadataResult:
        return PNGetAllChannelMetadataResult(envelope)

    def sync(self) -> PNGetAllChannelMetadataResultEnvelope:
        return PNGetAllChannelMetadataResultEnvelope(super().sync())

    def operation_type(self):
        return PNOperationType.PNGetAllChannelMetadataOperation

    def name(self):
        return "Get all Channels"

    def http_method(self):
        return HttpMethod.GET



from typing import List, Union
from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_CHANNELS_MISSING, PNERR_GROUP_MISSING
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.channel_group import PNChannelGroupsAddChannelResult
from pubnub.models.consumer.common import PNStatus
from pubnub.structures import Envelope


class PNChannelGroupsAddChannelResultEnvelope(Envelope):
    result: PNChannelGroupsAddChannelResult
    status: PNStatus


class AddChannelToChannelGroup(Endpoint):
    # /v1/channel-registration/sub-key/<sub_key>/channel-group/<group_name>?add=ch1,ch2
    ADD_PATH = "/v1/channel-registration/sub-key/%s/channel-group/%s"

    def __init__(self, pubnub, channels: Union[str, List[str]] = None, channel_group: str = None):
        Endpoint.__init__(self, pubnub)
        self._channels = []
        if channels:
            utils.extend_list(self._channels, channels)
        self._channel_group = channel_group

    def channels(self, channels) -> 'AddChannelToChannelGroup':
        utils.extend_list(self._channels, channels)
        return self

    def channel_group(self, channel_group: str) -> 'AddChannelToChannelGroup':
        self._channel_group = channel_group
        return self

    def custom_params(self):
        return {'add': utils.join_items(self._channels)}

    def build_path(self):
        return AddChannelToChannelGroup.ADD_PATH % (
            self.pubnub.config.subscribe_key, utils.url_encode(self._channel_group))

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()

        if len(self._channels) == 0:
            raise PubNubException(pn_error=PNERR_CHANNELS_MISSING)

        if not isinstance(self._channel_group, str) or len(self._channel_group) == 0:
            raise PubNubException(pn_error=PNERR_GROUP_MISSING)

    def is_auth_required(self):
        return True

    def create_response(self, envelope) -> PNChannelGroupsAddChannelResult:
        return PNChannelGroupsAddChannelResult()

    def sync(self) -> PNChannelGroupsAddChannelResultEnvelope:
        return PNChannelGroupsAddChannelResultEnvelope(super().sync())

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNAddChannelsToGroupOperation

    def name(self):
        return "AddChannelToChannelGroup"



from typing import List, Union
from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.presence import PNHereNowResult
from pubnub.structures import Envelope


class PNHereNowResultEnvelope(Envelope):
    result: PNHereNowResult
    status: PNStatus


class HereNow(Endpoint):
    HERE_NOW_PATH = "/v2/presence/sub-key/%s/channel/%s"
    HERE_NOW_GLOBAL_PATH = "/v2/presence/sub-key/%s"

    def __init__(self, pubnub, channels: Union[str, List[str]] = None, channel_groups: Union[str, List[str]] = None,
                 include_state: bool = False, include_uuids: bool = True):
        Endpoint.__init__(self, pubnub)
        self._channels = []
        if channels:
            utils.extend_list(self._channels, channels)

        self._channel_groups = []
        if channel_groups:
            utils.extend_list(self._channel_groups, channel_groups)

        self._include_state = include_state
        self._include_uuids = include_uuids

    def channels(self, channels: Union[str, List[str]]) -> 'HereNow':
        utils.extend_list(self._channels, channels)
        return self

    def channel_groups(self, channel_groups: Union[str, List[str]]) -> 'HereNow':
        utils.extend_list(self._channel_groups, channel_groups)
        return self

    def include_state(self, should_include_state) -> 'HereNow':
        self._include_state = should_include_state
        return self

    def include_uuids(self, include_uuids) -> 'HereNow':
        self._include_uuids = include_uuids
        return self

    def custom_params(self):
        params = {}

        if len(self._channel_groups) > 0:
            params['channel-group'] = utils.join_items_and_encode(self._channel_groups)

        if self._include_state:
            params['state'] = "1"

        if not self._include_uuids:
            params['disable_uuids'] = "1"

        return params

    def build_path(self):
        if len(self._channels) == 0 and len(self._channel_groups) == 0:
            return HereNow.HERE_NOW_GLOBAL_PATH % self.pubnub.config.subscribe_key
        else:
            return HereNow.HERE_NOW_PATH % (self.pubnub.config.subscribe_key,
                                            utils.join_channels(self._channels))

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()

    def is_auth_required(self):
        return True

    def create_response(self, envelope) -> PNHereNowResult:
        return PNHereNowResult.from_json(envelope, self._channels)

    def sync(self) -> PNHereNowResultEnvelope:
        return PNHereNowResultEnvelope(super().sync())

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNHereNowOperation

    def name(self):
        return "HereNow"



from typing import Optional
from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.signal import PNSignalResult
from pubnub.structures import Envelope


class SignalResultEnvelope(Envelope):
    result: PNSignalResult
    status: PNStatus


class Signal(Endpoint):
    SIGNAL_PATH = '/signal/%s/%s/0/%s/0/%s'

    _channel: str
    _message: any

    def __init__(self, pubnub, channel: str = None, message: any = None, custom_message_type: Optional[str] = None):
        Endpoint.__init__(self, pubnub)
        self._channel = channel
        self._message = message
        self._custom_message_type = custom_message_type

    def channel(self, channel) -> 'Signal':
        self._channel = str(channel)
        return self

    def message(self, message) -> 'Signal':
        self._message = message
        return self

    def custom_message_type(self, custom_message_type: str) -> 'Signal':
        self._custom_message_type = custom_message_type
        return self

    def build_path(self):
        stringified_message = utils.write_value_as_string(self._message)
        msg = utils.url_encode(stringified_message)
        return Signal.SIGNAL_PATH % (
            self.pubnub.config.publish_key, self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel), msg
        )

    def custom_params(self):
        params = {}
        if self._custom_message_type:
            params['custom_message_type'] = utils.url_encode(self._custom_message_type)

        return params

    def http_method(self):
        return HttpMethod.GET

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_publish_key()
        self.validate_channel()

    def create_response(self, result):  # pylint: disable=W0221
        return PNSignalResult(result)

    def sync(self) -> SignalResultEnvelope:
        return SignalResultEnvelope(super().sync())

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNSignalOperation

    def name(self):
        return "Signal"



import warnings
from typing import Any
from copy import deepcopy
from Cryptodome.Cipher import AES
from pubnub.enums import PNHeartbeatNotificationOptions, PNReconnectionPolicy
from pubnub.exceptions import PubNubException
from pubnub.crypto import PubNubCrypto, LegacyCryptoModule, PubNubCryptoModule


class PNConfiguration(object):
    DEFAULT_PRESENCE_TIMEOUT = 300
    DEFAULT_HEARTBEAT_INTERVAL = 280
    ALLOWED_AES_MODES = [AES.MODE_CBC, AES.MODE_GCM]
    DEFAULT_CRYPTO_MODULE = LegacyCryptoModule
    _locked = False

    def __init__(self):
        # TODO: add validation
        self._uuid = None
        self.origin = "ps.pndsn.com"
        self.ssl = True
        self.non_subscribe_request_timeout = 10
        self.subscribe_request_timeout = 310
        self.connect_timeout = 10
        self.subscribe_key = None
        self.publish_key = None
        self.secret_key = None
        self.cipher_key = None
        self._cipher_mode = AES.MODE_CBC
        self._fallback_cipher_mode = None
        self.auth_key = None
        self.filter_expression = None
        self.enable_subscribe = True
        self.crypto_instance = None
        self.file_crypto_instance = None
        self.log_verbosity = False
        self.enable_presence_heartbeat = False
        self.heartbeat_notification_options = PNHeartbeatNotificationOptions.FAILURES
        self.reconnect_policy = PNReconnectionPolicy.EXPONENTIAL
        self.maximum_reconnection_retries = None  # -1 means unlimited/ 0 means no retries
        self.reconnection_interval = None  # if None is left the default value from LinearDelay is used
        self.daemon = False
        self.use_random_initialization_vector = True
        self.suppress_leave_events = False
        self.should_compress = False

        self.heartbeat_default_values = True
        self._presence_timeout = PNConfiguration.DEFAULT_PRESENCE_TIMEOUT
        self._heartbeat_interval = PNConfiguration.DEFAULT_HEARTBEAT_INTERVAL
        self.cryptor = None
        self.file_cryptor = None
        self._crypto_module = None
        self.disable_config_locking = True
        self._locked = False

    def validate(self):
        PNConfiguration.validate_not_empty_string(self.uuid)
        if self.disable_config_locking:
            warnings.warn(DeprecationWarning('Mutable config will be deprecated in the future.'))

    def validate_not_empty_string(value: str):
        assert value and isinstance(value, str) and value.strip() != "", "UUID missing or invalid type"

    def scheme(self):
        if self.ssl:
            return "https"
        else:
            return "http"

    def scheme_extended(self):
        return self.scheme() + "://"

    def scheme_and_host(self):
        return self.scheme_extended() + self.origin

    def set_presence_timeout_with_custom_interval(self, timeout, interval):
        self.heartbeat_default_values = False
        self._presence_timeout = timeout
        self._heartbeat_interval = interval

    def set_presence_timeout(self, timeout):
        self.set_presence_timeout_with_custom_interval(timeout, (timeout / 2) - 1)

    @property
    def cipher_mode(self):
        return self._cipher_mode

    @cipher_mode.setter
    def cipher_mode(self, cipher_mode):
        if cipher_mode not in self.ALLOWED_AES_MODES:
            raise PubNubException('Cipher mode not supported')
        if cipher_mode is not self._cipher_mode:
            self._cipher_mode = cipher_mode
            self.crypto_instance = None

    @property
    def fallback_cipher_mode(self):
        return self._fallback_cipher_mode

    @fallback_cipher_mode.setter
    def fallback_cipher_mode(self, fallback_cipher_mode):
        if fallback_cipher_mode and fallback_cipher_mode not in self.ALLOWED_AES_MODES:
            raise PubNubException('Cipher mode not supported')
        if fallback_cipher_mode is not self._fallback_cipher_mode:
            self._fallback_cipher_mode = fallback_cipher_mode
            self.crypto_instance = None

    @property
    def crypto(self):
        if self._crypto_module:
            return self._crypto_module
        if self.crypto_instance is None:
            self._init_cryptodome()

        return self.crypto_instance

    def _init_cryptodome(self):
        if not self.cryptor:
            from pubnub.crypto import PubNubCryptodome
            self.cryptor = PubNubCryptodome
        self.crypto_instance = self.cryptor(self)

    def _init_file_crypto(self):
        from .crypto import PubNubFileCrypto
        if not self.file_cryptor:
            from pubnub.crypto import PubNubFileCrypto
            self.file_cryptor = PubNubFileCrypto
        self.file_crypto_instance = self.file_cryptor(self)

    @property
    def file_crypto(self) -> PubNubCrypto:
        if not self.file_crypto_instance:
            self._init_file_crypto()

        return self.file_crypto_instance

    @property
    def crypto_module(self):
        return self._crypto_module

    @crypto_module.setter
    def crypto_module(self, crypto_module: PubNubCryptoModule):
        self._crypto_module = crypto_module

    @property
    def port(self):
        return 443 if self.ssl == "https" else 80

    @property
    def presence_timeout(self):
        return self._presence_timeout

    @property
    def heartbeat_interval(self):
        return self._heartbeat_interval

        # TODO: set log level
        # TODO: set log level

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, uuid):
        PNConfiguration.validate_not_empty_string(uuid)
        self._uuid = uuid

    @property
    def user_id(self):
        return self._uuid

    @user_id.setter
    def user_id(self, user_id):
        PNConfiguration.validate_not_empty_string(user_id)
        self._uuid = user_id

    def lock(self):
        self.__dict__['_locked'] = False if self.disable_config_locking else True

    def copy(self):
        config_copy = deepcopy(self)
        config_copy.__dict__['_locked'] = False
        return config_copy

    def __setattr__(self, name: str, value: Any) -> None:
        if self._locked:
            warnings.warn(UserWarning('Configuration is locked. Any changes made won\'t have any effect'))
            return
        if name in ['uuid', 'user_id']:
            PNConfiguration.validate_not_empty_string(value)
            self.__dict__['_uuid'] = value
        elif name in ['cipher_mode', 'fallback_cipher_mode', 'crypto_module']:
            self.__dict__[f'_{name}'] = value
        else:
            self.__dict__[name] = value



from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_PUSH_DEVICE_MISSING, PNERROR_PUSH_TYPE_MISSING, PNERR_PUSH_TOPIC_MISSING
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType, PNPushType, PNPushEnvironment
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.push import PNPushRemoveAllChannelsResult
from pubnub import utils
from pubnub.structures import Envelope


class PNPushRemoveAllChannelsResultEnvelope(Envelope):
    result: PNPushRemoveAllChannelsResult
    status: PNStatus


class RemoveDeviceFromPush(Endpoint):
    # v1/push/sub-key/{subKey}/devices/{pushToken}/remove
    REMOVE_PATH = "/v1/push/sub-key/%s/devices/%s/remove"
    # v2/push/sub-key/{subKey}/devices-apns2/{deviceApns2}/remove
    REMOVE_PATH_APNS2 = "/v2/push/sub-key/%s/devices-apns2/%s/remove"

    def __init__(self, pubnub, device_id: str = None, push_type: PNPushType = None, topic: str = None,
                 environment: PNPushEnvironment = None):
        Endpoint.__init__(self, pubnub)
        self._device_id = device_id
        self._push_type = push_type
        self._topic = topic
        self._environment = environment

    def device_id(self, device_id: str) -> 'RemoveDeviceFromPush':
        self._device_id = device_id
        return self

    def push_type(self, push_type: PNPushType) -> 'RemoveDeviceFromPush':
        self._push_type = push_type
        return self

    def topic(self, topic: str) -> 'RemoveDeviceFromPush':
        self._topic = topic
        return self

    def environment(self, environment: PNPushEnvironment) -> 'RemoveDeviceFromPush':
        self._environment = environment
        return self

    def custom_params(self):
        params = {}

        if self._push_type != PNPushType.APNS2:
            params['type'] = utils.push_type_to_string(self._push_type)
        else:
            if self._environment is None:
                self._environment = PNPushEnvironment.DEVELOPMENT

            params['environment'] = self._environment
            params['topic'] = self._topic

        return params

    def build_path(self):
        if self._push_type != PNPushType.APNS2:
            return RemoveDeviceFromPush.REMOVE_PATH % (
                self.pubnub.config.subscribe_key, self._device_id)
        else:
            return RemoveDeviceFromPush.REMOVE_PATH_APNS2 % (
                self.pubnub.config.subscribe_key, self._device_id)

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()

        if not isinstance(self._device_id, str) or len(self._device_id) == 0:
            raise PubNubException(pn_error=PNERR_PUSH_DEVICE_MISSING)

        if self._push_type is None:
            raise PubNubException(pn_error=PNERROR_PUSH_TYPE_MISSING)

        if self._push_type == PNPushType.APNS2:
            if not isinstance(self._topic, str) or len(self._topic) == 0:
                raise PubNubException(pn_error=PNERR_PUSH_TOPIC_MISSING)

    def create_response(self, envelope) -> PNPushRemoveAllChannelsResult:
        return PNPushRemoveAllChannelsResult()

    def sync(self) -> PNPushRemoveAllChannelsResultEnvelope:
        return PNPushRemoveAllChannelsResultEnvelope(super().sync())

    def is_auth_required(self):
        return True

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNRemoveAllPushNotificationsOperation

    def name(self):
        return "RemoveDeviceFromPush"



from typing import List, Union
from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.presence import PNGetStateResult
from pubnub.endpoints.mixins import UUIDValidatorMixin
from pubnub.structures import Envelope


class PNGetStateResultEnvelope(Envelope):
    result: PNGetStateResult
    status: PNStatus


class GetState(Endpoint, UUIDValidatorMixin):
    # /v2/presence/sub-key/<subscribe_key>/channel/<channel>/uuid/<uuid>/data?state=<state>
    GET_STATE_PATH = "/v2/presence/sub-key/%s/channel/%s/uuid/%s"

    def __init__(self, pubnub, channels: Union[str, List[str]] = None, channel_groups: Union[str, List[str]] = None,
                 uuid: str = None):
        Endpoint.__init__(self, pubnub)
        self._channels = []
        if channels:
            utils.extend_list(self._channels, channels)

        self._groups = []
        if channel_groups:
            utils.extend_list(self._groups, channel_groups)

        self._uuid = self.pubnub.uuid
        if uuid:
            self._uuid = uuid

    def channels(self, channels: Union[str, List[str]]) -> 'GetState':
        utils.extend_list(self._channels, channels)
        return self

    def uuid(self, uuid: str) -> 'GetState':
        self._uuid = uuid
        return self

    def channel_groups(self, channel_groups: Union[str, List[str]]) -> 'GetState':
        utils.extend_list(self._groups, channel_groups)
        return self

    def custom_params(self):
        params = {}

        if len(self._groups) > 0:
            params['channel-group'] = utils.join_items(self._groups)

        return params

    def build_path(self):
        return GetState.GET_STATE_PATH % (
            self.pubnub.config.subscribe_key,
            utils.join_channels(self._channels),
            utils.url_encode(self._uuid)
        )

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channels_and_groups()
        self.validate_uuid()

    def create_response(self, envelope) -> PNGetStateResult:
        if len(self._channels) == 1 and len(self._groups) == 0:
            channels = {self._channels[0]: envelope['payload']}
        else:
            channels = envelope['payload']['channels']

        return PNGetStateResult(channels)

    def sync(self) -> PNGetStateResultEnvelope:
        return PNGetStateResultEnvelope(super().sync())

    def is_auth_required(self):
        return True

    def affected_channels(self):
        return self._channels

    def affected_channels_groups(self):
        return self._groups

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNGetState

    def name(self):
        return "GetState"



from abc import ABCMeta, abstractmethod

from pubnub.models.subscription import PubNubChannel, PubNubChannelGroup
from . import utils


class PubSubBuilder(object):
    __metaclass__ = ABCMeta

    def __init__(self, pubnub_instance):
        self._pubnub = pubnub_instance
        self._channel_subscriptions = []
        self._channel_group_subscriptions = []

    # TODO: make the 'channel' alias
    def channels(self, channels_list):
        utils.extend_list(self._channel_subscriptions, channels_list)

        return self

    def channel_groups(self, channel_groups_list):
        utils.extend_list(self._channel_group_subscriptions, channel_groups_list)

        return self

    @abstractmethod
    def execute(self):
        pass


class SubscribeBuilder(PubSubBuilder):
    def __init__(self, pubnub_instance):
        super(SubscribeBuilder, self).__init__(pubnub_instance)
        self._presence_enabled = False
        self._timetoken = 0

    def with_presence(self):
        self._presence_enabled = True
        return self

    def with_timetoken(self, timetoken):
        self._timetoken = timetoken
        return self

    def channel_subscriptions(self):
        return [PubNubChannel(self._pubnub, channel).subscription(self._presence_enabled)
                for channel in self._channel_subscriptions]

    def channel_group_subscriptions(self):
        return [PubNubChannelGroup(self._pubnub, group).subscription(self._presence_enabled)
                for group in self._channel_group_subscriptions]

    def execute(self):
        subscription = self._pubnub.subscription_set(self.channel_subscriptions() + self.channel_group_subscriptions())

        subscription.subscribe(timetoken=self._timetoken)


class UnsubscribeBuilder(PubSubBuilder):
    def execute(self):
        self._pubnub._subscription_registry.unsubscribe(channels=self._channel_subscriptions,
                                                        groups=self._channel_group_subscriptions)



from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, UuidEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.objects_v2.uuid import PNRemoveUUIDMetadataResult
from pubnub.structures import Envelope


class PNRemoveUUIDMetadataResultEnvelope(Envelope):
    result: PNRemoveUUIDMetadataResult
    status: PNStatus


class RemoveUuid(ObjectsEndpoint, UuidEndpoint):
    REMOVE_UID_PATH = "/v2/objects/%s/uuids/%s"

    def __init__(self, pubnub, uuid: str = None):
        ObjectsEndpoint.__init__(self, pubnub)
        UuidEndpoint.__init__(self, uuid=uuid)

    def build_path(self):
        return RemoveUuid.REMOVE_UID_PATH % (self.pubnub.config.subscribe_key, self._effective_uuid())

    def validate_specific_params(self):
        self._validate_uuid()

    def create_response(self, envelope) -> PNRemoveUUIDMetadataResult:
        return PNRemoveUUIDMetadataResult(envelope)

    def sync(self) -> PNRemoveUUIDMetadataResultEnvelope:
        return PNRemoveUUIDMetadataResultEnvelope(super().sync())

    def operation_type(self):
        return PNOperationType.PNRemoveUuidMetadataOperation

    def name(self):
        return "Remove UUID"

    def http_method(self):
        return HttpMethod.DELETE



from pubnub import utils
from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, UuidEndpoint, \
    IncludeCustomEndpoint, CustomAwareEndpoint, IncludeStatusTypeEndpoint, StatusTypeAwareEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.objects_v2.uuid import PNSetUUIDMetadataResult
from pubnub.structures import Envelope


class PNSetUUIDMetadataResultEnvelope(Envelope):
    result: PNSetUUIDMetadataResult
    status: PNStatus


class SetUuid(ObjectsEndpoint, UuidEndpoint, IncludeCustomEndpoint, CustomAwareEndpoint, IncludeStatusTypeEndpoint,
              StatusTypeAwareEndpoint):
    SET_UID_PATH = "/v2/objects/%s/uuids/%s"

    def __init__(self, pubnub, uuid: str = None, include_custom: bool = None, custom: dict = None,
                 include_status: bool = True, include_type: bool = True, status: str = None, type: str = None,
                 name: str = None, email: str = None, external_id: str = None, profile_url: str = None):
        ObjectsEndpoint.__init__(self, pubnub)
        UuidEndpoint.__init__(self, uuid=uuid)
        IncludeCustomEndpoint.__init__(self, include_custom=include_custom)
        CustomAwareEndpoint.__init__(self, custom=custom)
        IncludeStatusTypeEndpoint.__init__(self, include_status=include_status, include_type=include_type)
        StatusTypeAwareEndpoint.__init__(self, status=status, type=type)

        self._name = name
        self._email = email
        self._external_id = external_id
        self._profile_url = profile_url

    def set_name(self, name: str):
        self._name = str(name)
        return self

    def email(self, email: str):
        self._email = str(email)
        return self

    def external_id(self, external_id: str):
        self._external_id = str(external_id)
        return self

    def profile_url(self, profile_url: str):
        self._profile_url = str(profile_url)
        return self

    def build_path(self):
        return SetUuid.SET_UID_PATH % (self.pubnub.config.subscribe_key, self._effective_uuid())

    def build_data(self):
        payload = {
            "name": self._name,
            "email": self._email,
            "externalId": self._external_id,
            "profileUrl": self._profile_url,
            "custom": self._custom
        }
        payload = StatusTypeAwareEndpoint.build_data(self, payload)
        return utils.write_value_as_string(payload)

    def validate_specific_params(self):
        self._validate_uuid()

    def create_response(self, envelope) -> PNSetUUIDMetadataResult:
        return PNSetUUIDMetadataResult(envelope)

    def sync(self) -> PNSetUUIDMetadataResultEnvelope:
        return PNSetUUIDMetadataResultEnvelope(super().sync())

    def operation_type(self):
        return PNOperationType.PNSetUuidMetadataOperation

    def name(self):
        return "Set UUID"

    def http_method(self):
        return HttpMethod.PATCH



from enum import Enum
from typing import List, Optional, Union

from pubnub.callbacks import SubscribeCallback
from pubnub.dtos import SubscribeOperation, UnsubscribeOperation


class PNSubscriptionType(Enum):
    CHANNEL: str = "channel"
    CHANNEL_GROUP: str = "channel_group"


class PNSubscribable:
    pubnub = None
    name: str
    _type: PNSubscriptionType = None

    def __init__(self, pubnub_instance, name) -> None:
        self.pubnub = pubnub_instance
        self.name = name

    def subscription(self, with_presence: bool = None):
        return PubNubSubscription(self.pubnub, self.name, self._type, with_presence=with_presence)


class PNEventEmitter:
    on_message: callable
    on_signal: callable
    on_presence: callable
    on_channel_metadata: callable
    on_user_metadata: callable
    on_message_action: callable
    on_membership: callable
    on_file: callable

    def is_matching_listener(self, message):
        def wildcard_match(name, subscription):
            return subscription.endswith('.*') and name.startswith(subscription.strip('*'))
        if isinstance(self, PubNubSubscriptionSet):
            return any([subscription_item.is_matching_listener(message)
                        for subscription_item in self.get_subscription_items()])
        else:
            if self._type == PNSubscriptionType.CHANNEL:
                return message.channel == self.name or wildcard_match(message.channel, self.name)
            else:
                return message.subscription == self.name

    def presence(self, presence):
        if not hasattr(self, 'on_presence') or not (hasattr(self, 'with_presence') and self.with_presence):
            return

        if self.is_matching_listener(presence) and hasattr(self, 'on_presence'):
            self.on_presence(presence)

    def message(self, message):
        if self.is_matching_listener(message) and hasattr(self, 'on_message'):
            self.on_message(message)

    def message_action(self, message_action):
        if self.is_matching_listener(message_action) and hasattr(self, 'on_message_action'):
            self.on_message_action(message_action)

    def signal(self, signal):
        if self.is_matching_listener(signal) and hasattr(self, 'on_signal'):
            self.on_signal(signal)


class PNSubscribeCapable:
    def subscribe(self, timetoken: Optional[int] = None, region: Optional[str] = None):
        self.timetoken = timetoken
        self.region = region
        self.subscription_registry.add(self)

    def unsubscribe(self):
        self.subscription_registry.remove(self)


class PubNubSubscription(PNEventEmitter, PNSubscribeCapable):
    def __init__(self, pubnub_instance, name: str, type: PNSubscriptionType, with_presence: bool = False) -> None:
        self.subscription_registry = pubnub_instance._subscription_registry
        self.subscription_manager = pubnub_instance._subscription_manager
        self.name = name
        self._type = type
        self.with_presence = with_presence

    def add_listener(self, listener):
        self.subscription_registry.add_listener(listener)

    def get_names_with_presence(self):
        return [self.name, f'{self.name}-pnpres'] if self.with_presence else [self.name]


class PubNubSubscriptionSet(PNEventEmitter, PNSubscribeCapable):
    def __init__(self, pubnub_instance, subscriptions: List[PubNubSubscription]) -> None:
        self.subscription_registry = pubnub_instance._subscription_registry
        self.subscription_manager = pubnub_instance._subscription_manager
        self.subscriptions = subscriptions

    def get_subscription_items(self):
        return [item for item in self.subscriptions]


class PubNubChannel(PNSubscribable):
    _type = PNSubscriptionType.CHANNEL

    def __init__(self, pubnub_instance, channel: str) -> None:
        super().__init__(pubnub_instance, channel)


class PubNubChannelGroup(PNSubscribable):
    _type = PNSubscriptionType.CHANNEL_GROUP

    def __init__(self, pubnub_instance, channel_group: str) -> None:
        super().__init__(pubnub_instance, channel_group)


class PubNubChannelMetadata(PNSubscribable):
    _type = PNSubscriptionType.CHANNEL

    def __init__(self, pubnub_instance, channel: str) -> None:
        super().__init__(pubnub_instance, channel)


class PubNubUserMetadata(PNSubscribable):
    _types = PNSubscriptionType.CHANNEL

    def __init__(self, pubnub_instance, user_id: str) -> None:
        super().__init__(pubnub_instance, user_id)


class PNSubscriptionRegistry:
    def __init__(self, pubnub_instance):
        self.pubnub = pubnub_instance
        self.global_listeners = []
        self.channels = {}
        self.channel_groups = {}
        self.subscription_registry_callback = None
        self.with_presence = None
        self.subscriptions = []

    def __add_subscription(self, subscription: PubNubSubscription, subscription_set: PubNubSubscriptionSet = None):
        names_added = []
        self.subscriptions.append(subscription)

        subscriptions = [subscription]
        if subscription_set:
            subscriptions.append(subscription_set)

        if subscription._type == PNSubscriptionType.CHANNEL:
            subscription_list = self.channels
        else:
            subscription_list = self.channel_groups

        for name in subscription.get_names_with_presence():
            if name not in subscription_list:
                subscription_list[name] = subscriptions
                names_added.append(name)
            else:
                subscription_list[name].extend(subscriptions)
        return names_added

    def __remove_subscription(self, subscription: PubNubSubscription):
        names_removed = {'channels': [],
                         'groups': []}

        self.subscriptions.remove(subscription)

        if subscription._type == PNSubscriptionType.CHANNEL:
            subscription_list = self.channels
            removed = names_removed['channels']
        else:
            subscription_list = self.channel_groups
            removed = names_removed['groups']

        for name in subscription.get_names_with_presence():
            if name in subscription_list and subscription in subscription_list[name]:
                subscription_list[name].remove(subscription)
                if len(subscription_list[name]) == 0:
                    removed.append(name)

        return names_removed

    def add(self, subscription: Union[PubNubSubscription, PubNubSubscriptionSet]) -> list:
        if not self.subscription_registry_callback:
            self.subscription_registry_callback = PNSubscriptionRegistryCallback(self)
            self.pubnub.add_listener(self.subscription_registry_callback)

        self.with_presence = any(sub.with_presence for sub in self.subscriptions)

        names_changed = []
        if isinstance(subscription, PubNubSubscriptionSet):
            for subscription_part in subscription.subscriptions:
                names_changed.append(self.__add_subscription(subscription_part, subscription))
        else:
            names_changed.append(self.__add_subscription(subscription))

        tt = self.pubnub._subscription_manager._timetoken
        if subscription.timetoken:
            tt = max(subscription.timetoken, self.pubnub._subscription_manager._timetoken)

        if names_changed:
            subscribe_operation = SubscribeOperation(
                channels=self.get_subscribed_channels(),
                channel_groups=self.get_subscribed_channel_groups(),
                timetoken=tt,
                presence_enabled=self.with_presence,
            )
            self.pubnub._subscription_manager.adapt_subscribe_builder(subscribe_operation)
        return names_changed

    def remove(self, subscription: Union[PubNubSubscription, PubNubSubscriptionSet]) -> list:
        channels_changed = []
        groups_changed = []

        if isinstance(subscription, PubNubSubscriptionSet):
            for subscription_part in subscription.subscriptions:
                names_changed = self.__remove_subscription(subscription_part)
                channels_changed += names_changed['channels']
                groups_changed += names_changed['groups']
        else:
            names_changed = self.__remove_subscription(subscription)
            channels_changed += names_changed['channels']
            groups_changed += names_changed['groups']

        self.with_presence = any(sub.with_presence for sub in self.subscriptions)

        if names_changed:
            unsubscribe_operation = UnsubscribeOperation(channels=channels_changed, channel_groups=groups_changed)
            self.pubnub._subscription_manager.adapt_unsubscribe_builder(unsubscribe_operation)
        return names_changed

    def get_subscribed_channels(self):
        return list(self.channels.keys())

    def get_subscribed_channel_groups(self):
        return list(self.channel_groups.keys())

    def get_subscriptions_for(self, _type: PNSubscriptionType, name: str):
        if _type == PNSubscriptionType.CHANNEL:
            return [channel for channel in self.get_subscribed_channels() if channel == name]
        else:
            return [group for group in self.get_subscribed_channel_groups() if group == name]

    def get_all_listeners(self):
        listeners = []

        for channel in self.channels:
            listeners += self.channels[channel]
        for channel_group in self.channel_groups:
            listeners += self.channel_groups[channel_group]
        if self.global_listeners:
            listeners += self.global_listeners
        return set(listeners)

    def add_listener(self, listener):
        assert isinstance(listener, SubscribeCallback)
        self.global_listeners.append(listener)

    def remove_listener(self, listener):
        assert isinstance(listener, SubscribeCallback)
        self.global_listeners.remove(listener)

    def unsubscribe_all(self):
        unsubscribe_operation = UnsubscribeOperation(
            channels=list(self.channels.keys()),
            channel_groups=list(self.channel_groups.keys())
        )
        self.pubnub._subscription_manager.adapt_unsubscribe_builder(unsubscribe_operation)
        self.channels = []
        self.channel_groups = []

    def unsubscribe(self, channels=None, groups=None):
        presence_channels = []
        for channel in channels:
            del self.channels[channel]
            if f'{channel}-pnpres' in self.channels:
                del self.channels[f'{channel}-pnpres']
                presence_channels.append(f'{channel}-pnpres')

        presence_groups = []
        for group in groups:
            del self.channel_groups[group]
            if f'{group}-pnpres' in self.channel_groups:
                del self.channel_groups[f'{group}-pnpres']
                presence_groups.append(f'{group}-pnpres')

        unsubscribe_operation = UnsubscribeOperation(
            channels=channels + presence_channels,
            channel_groups=groups + presence_groups
        )
        self.pubnub._subscription_manager.adapt_unsubscribe_builder(unsubscribe_operation)


class PNSubscriptionRegistryCallback(SubscribeCallback):
    def __init__(self, subscription_registry: PNSubscriptionRegistry) -> None:
        self.subscription_registry = subscription_registry
        super().__init__()

    def status(self, _, status):
        pass

    def presence(self, _, presence):
        for listener in self.subscription_registry.get_all_listeners():
            listener.presence(presence)

    def message(self, _, message):
        for listener in self.subscription_registry.get_all_listeners():
            listener.message(message)

    def signal(self, _, signal):
        for listener in self.subscription_registry.get_all_listeners():
            listener.signal(signal)

    def channel(self, _, channel):
        for listener in self.subscription_registry.get_all_listeners():
            listener.channel(channel)

    def uuid(self, pubnub, uuid):
        for listener in self.subscription_registry.get_all_listeners():
            listener.uuid(uuid)

    def membership(self, _, membership):
        for listener in self.subscription_registry.get_all_listeners():
            listener.membership(membership)

    def message_action(self, _, message_action):
        for listener in self.subscription_registry.get_all_listeners():
            listener.message_action(message_action)

    def file(self, _, file_message):
        for listener in self.subscription_registry.get_all_listeners():
            listener.file_message(file_message)



from typing import List
from pubnub import utils
from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, ChannelEndpoint, ListEndpoint, \
    IncludeCustomEndpoint, UUIDIncludeEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.objects_v2.channel_members import PNRemoveChannelMembersResult
from pubnub.models.consumer.objects_v2.page import PNPage
from pubnub.structures import Envelope


class PNRemoveChannelMembersResultEnvelope(Envelope):
    result: PNRemoveChannelMembersResult
    status: PNStatus


class RemoveChannelMembers(ObjectsEndpoint, ChannelEndpoint, ListEndpoint, IncludeCustomEndpoint,
                           UUIDIncludeEndpoint):
    REMOVE_CHANNEL_MEMBERS_PATH = "/v2/objects/%s/channels/%s/uuids"

    def __init__(self, pubnub, channel: str = None, uuids: List[str] = None, include_custom: bool = None,
                 limit: int = None, filter: str = None, include_total_count: bool = None, sort_keys: list = None,
                 page: PNPage = None):
        ObjectsEndpoint.__init__(self, pubnub)
        ListEndpoint.__init__(self, limit=limit, filter=filter, include_total_count=include_total_count,
                              sort_keys=sort_keys, page=page)
        ChannelEndpoint.__init__(self, channel=channel)
        IncludeCustomEndpoint.__init__(self, include_custom=include_custom)
        UUIDIncludeEndpoint.__init__(self)

        self._uuids = []
        if uuids:
            utils.extend_list(self._uuids, uuids)

    def uuids(self, uuids: List[str]) -> 'RemoveChannelMembers':
        utils.extend_list(self._uuids, uuids)
        return self

    def build_path(self):
        return RemoveChannelMembers.REMOVE_CHANNEL_MEMBERS_PATH % (self.pubnub.config.subscribe_key, self._channel)

    def build_data(self):
        uuids_to_delete = []

        for uuid in self._uuids:
            uuids_to_delete.append(uuid.to_payload_dict())

        payload = {
            "set": [],
            "delete": uuids_to_delete
        }
        return utils.write_value_as_string(payload)

    def validate_specific_params(self):
        self._validate_channel()

    def create_response(self, envelope) -> PNRemoveChannelMembersResult:
        return PNRemoveChannelMembersResult(envelope)

    def sync(self) -> PNRemoveChannelMembersResultEnvelope:
        return PNRemoveChannelMembersResultEnvelope(super().sync())

    def operation_type(self):
        return PNOperationType.PNRemoveChannelMembersOperation

    def name(self):
        return "Remove Channel Members"

    def http_method(self):
        return HttpMethod.PATCH



from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, IncludeCustomEndpoint, ChannelEndpoint, \
    IncludeStatusTypeEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.objects_v2.channel import PNGetChannelMetadataResult
from pubnub.structures import Envelope


class PNGetChannelMetadataResultEnvelope(Envelope):
    result: PNGetChannelMetadataResult
    status: PNStatus


class GetChannel(ObjectsEndpoint, ChannelEndpoint, IncludeCustomEndpoint, IncludeStatusTypeEndpoint):
    GET_CHANNEL_PATH = "/v2/objects/%s/channels/%s"

    def __init__(self, pubnub, channel: str = None, include_custom: bool = False, include_status: bool = True,
                 include_type: bool = True):
        ObjectsEndpoint.__init__(self, pubnub)
        ChannelEndpoint.__init__(self, channel=channel)
        IncludeCustomEndpoint.__init__(self, include_custom=include_custom)
        IncludeStatusTypeEndpoint.__init__(self, include_status=include_status, include_type=include_type)

    def build_path(self):
        return GetChannel.GET_CHANNEL_PATH % (self.pubnub.config.subscribe_key, self._channel)

    def validate_specific_params(self):
        self._validate_channel()

    def create_response(self, envelope) -> PNGetChannelMetadataResult:
        return PNGetChannelMetadataResult(envelope)

    def sync(self) -> PNGetChannelMetadataResultEnvelope:
        return PNGetChannelMetadataResultEnvelope(super().sync())

    def operation_type(self):
        return PNOperationType.PNGetChannelMetadataOperation

    def name(self):
        return "Get Channel"

    def http_method(self):
        return HttpMethod.GET



from pubnub import utils
from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, IncludeCustomEndpoint, \
    ChannelEndpoint, CustomAwareEndpoint, IncludeStatusTypeEndpoint, StatusTypeAwareEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.objects_v2.channel import PNSetChannelMetadataResult
from pubnub.structures import Envelope


class PNSetChannelMetadataResultEnvelope(Envelope):
    result: PNSetChannelMetadataResult
    status: PNStatus


class SetChannel(ObjectsEndpoint, ChannelEndpoint, IncludeCustomEndpoint, CustomAwareEndpoint,
                 IncludeStatusTypeEndpoint, StatusTypeAwareEndpoint):
    SET_CHANNEL_PATH = "/v2/objects/%s/channels/%s"

    def __init__(self, pubnub, channel: str = None, custom: dict = None, include_custom: bool = False,
                 include_status: bool = True, include_type: bool = True, name: str = None, description: str = None,
                 status: str = None, type: str = None):
        ObjectsEndpoint.__init__(self, pubnub)
        ChannelEndpoint.__init__(self, channel=channel)
        CustomAwareEndpoint.__init__(self, custom=custom)
        IncludeCustomEndpoint.__init__(self, include_custom=include_custom)
        IncludeStatusTypeEndpoint.__init__(self, include_status=include_status, include_type=include_type)
        self._name = name
        self._description = description
        self._status = status
        self._type = type

    def set_name(self, name: str) -> 'SetChannel':
        self._name = str(name)
        return self

    def set_status(self, status: str = None) -> 'SetChannel':
        self._status = status
        return self

    def set_type(self, type: str = None) -> 'SetChannel':
        self._type = type
        return self

    def description(self, description) -> 'SetChannel':
        self._description = str(description)
        return self

    def validate_specific_params(self):
        self._validate_channel()

    def build_path(self):
        return SetChannel.SET_CHANNEL_PATH % (self.pubnub.config.subscribe_key, self._channel)

    def build_data(self):
        payload = {
            "name": self._name,
            "description": self._description,
            "status": self._status,
            "type": self._type,
            "custom": self._custom
        }
        payload = StatusTypeAwareEndpoint.build_data(self, payload)
        return utils.write_value_as_string(payload)

    def create_response(self, envelope) -> PNSetChannelMetadataResult:
        return PNSetChannelMetadataResult(envelope)

    def sync(self) -> PNSetChannelMetadataResultEnvelope:
        return PNSetChannelMetadataResultEnvelope(super().sync())

    def operation_type(self):
        return PNOperationType.PNSetChannelMetadataOperation

    def name(self):
        return "Set UUID"

    def http_method(self):
        return HttpMethod.PATCH



from typing import Union, List, Optional
from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_TTL_MISSING, PNERR_INVALID_META, PNERR_RESOURCES_MISSING
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.v3.access_manager import PNGrantTokenResult
from pubnub.structures import Envelope


class PNGrantTokenResultEnvelope(Envelope):
    result: PNGrantTokenResult
    status: PNStatus


class GrantToken(Endpoint):
    GRANT_TOKEN_PATH = "/v3/pam/%s/grant"

    def __init__(self, pubnub, channels: Union[str, List[str]] = None, channel_groups: Union[str, List[str]] = None,
                 users: Union[str, List[str]] = None, spaces: Union[str, List[str]] = None,
                 authorized_user_id: str = None, ttl: Optional[int] = None, meta: Optional[any] = None):
        Endpoint.__init__(self, pubnub)
        self._ttl = ttl
        self._meta = meta
        self._authorized_uuid = authorized_user_id
        self._channels = []
        if channels:
            utils.extend_list(self._channels, channels)
        if spaces:
            utils.extend_list(self._channels, spaces)

        self._groups = []
        if channel_groups:
            utils.extend_list(self._groups, channel_groups)
        self._uuids = []
        if users:
            utils.extend_list(self._uuids, users)

        self._sort_params = True

    def ttl(self, ttl: int) -> 'GrantToken':
        self._ttl = ttl
        return self

    def meta(self, meta: any) -> 'GrantToken':
        self._meta = meta
        return self

    def authorized_uuid(self, uuid: str) -> 'GrantToken':
        self._authorized_uuid = uuid
        return self

    def authorized_user(self, user) -> 'GrantToken':
        self._authorized_uuid = user
        return self

    def spaces(self, spaces: Union[str, List[str]]) -> 'GrantToken':
        self._channels = spaces
        return self

    def users(self, users: Union[str, List[str]]) -> 'GrantToken':
        self._uuids = users
        return self

    def channels(self, channels: Union[str, List[str]]) -> 'GrantToken':
        self._channels = channels
        return self

    def groups(self, groups: Union[str, List[str]]) -> 'GrantToken':
        self._groups = groups
        return self

    def uuids(self, uuids: Union[str, List[str]]) -> 'GrantToken':
        self._uuids = uuids
        return self

    def custom_params(self):
        return {}

    def build_data(self):
        params = {'ttl': int(self._ttl)}

        permissions = {}
        resources = {}
        patterns = {}

        utils.parse_resources(self._channels, "channels", resources, patterns)
        utils.parse_resources(self._groups, "groups", resources, patterns)
        utils.parse_resources(self._uuids, "uuids", resources, patterns)
        utils.parse_resources(self._uuids, "users", resources, patterns)
        utils.parse_resources(self._channels, "spaces", resources, patterns)

        permissions['resources'] = resources
        permissions['patterns'] = patterns

        if self._meta:
            if isinstance(self._meta, dict):
                permissions['meta'] = self._meta
            else:
                raise PubNubException(pn_error=PNERR_INVALID_META)
        else:
            permissions['meta'] = {}

        if self._authorized_uuid:
            permissions["uuid"] = self._authorized_uuid

        params['permissions'] = permissions

        return utils.write_value_as_string(params)

    def build_path(self):
        return GrantToken.GRANT_TOKEN_PATH % self.pubnub.config.subscribe_key

    def http_method(self):
        return HttpMethod.POST

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_secret_key()
        self.validate_ttl()
        self.validate_resources()

    def create_response(self, envelope) -> PNGrantTokenResult:
        return PNGrantTokenResult.from_json(envelope['data'])

    def sync(self) -> PNGrantTokenResultEnvelope:
        return PNGrantTokenResultEnvelope(super().sync())

    def is_auth_required(self):
        return False

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNAccessManagerGrantToken

    def name(self):
        return "Grant Token"

    def validate_resources(self):
        if not any((self._channels, self._groups, self._uuids)):
            raise PubNubException(pn_error=PNERR_RESOURCES_MISSING)

    def validate_ttl(self):
        if not self._ttl:
            raise PubNubException(pn_error=PNERR_TTL_MISSING)



from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, IncludeCustomEndpoint, \
    ChannelEndpoint, ListEndpoint, UUIDIncludeEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.objects_v2.channel_members import PNGetChannelMembersResult
from pubnub.models.consumer.objects_v2.page import PNPage
from pubnub.structures import Envelope


class PNGetChannelMembersResultEnvelope(Envelope):
    result: PNGetChannelMembersResult
    status: PNStatus


class GetChannelMembers(ObjectsEndpoint, ChannelEndpoint, ListEndpoint, IncludeCustomEndpoint,
                        UUIDIncludeEndpoint):
    GET_CHANNEL_MEMBERS_PATH = "/v2/objects/%s/channels/%s/uuids"

    def __init__(self, pubnub, channel: str = None, include_custom: bool = None, limit: int = None, filter: str = None,
                 include_total_count: bool = None, sort_keys: list = None, page: PNPage = None):
        ObjectsEndpoint.__init__(self, pubnub)
        ChannelEndpoint.__init__(self, channel=channel)
        ListEndpoint.__init__(self, limit=limit, filter=filter, include_total_count=include_total_count,
                              sort_keys=sort_keys, page=page)
        IncludeCustomEndpoint.__init__(self, include_custom=include_custom)
        UUIDIncludeEndpoint.__init__(self)

    def build_path(self):
        return GetChannelMembers.GET_CHANNEL_MEMBERS_PATH % (self.pubnub.config.subscribe_key, self._channel)

    def validate_specific_params(self):
        self._validate_channel()

    def create_response(self, envelope) -> PNGetChannelMembersResult:
        return PNGetChannelMembersResult(envelope)

    def sync(self) -> PNGetChannelMembersResultEnvelope:
        return PNGetChannelMembersResultEnvelope(super().sync())

    def operation_type(self):
        return PNOperationType.PNGetChannelMembersOperation

    def name(self):
        return "Get Channel Members"

    def http_method(self):
        return HttpMethod.GET



from typing import Optional
from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.presence import PNWhereNowResult
from pubnub.endpoints.mixins import UUIDValidatorMixin
from pubnub.structures import Envelope


class PNWhereNowResultEnvelope(Envelope):
    result: PNWhereNowResult
    status: PNStatus


class WhereNow(Endpoint, UUIDValidatorMixin):
    # /v2/presence/sub-key/<subscribe_key>/uuid/<uuid>
    WHERE_NOW_PATH = "/v2/presence/sub-key/%s/uuid/%s"

    def __init__(self, pubnub, uuid: Optional[str] = None):
        Endpoint.__init__(self, pubnub)
        self._uuid = pubnub.config.uuid
        if uuid:
            self._uuid = uuid

    def uuid(self, uuid: str) -> 'WhereNow':
        self._uuid = uuid
        return self

    def custom_params(self):
        return {}

    def build_path(self):
        return WhereNow.WHERE_NOW_PATH % (self.pubnub.config.subscribe_key, utils.url_encode(self._uuid))

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_uuid()

    def is_auth_required(self):
        return True

    def create_response(self, envelope):
        return PNWhereNowResult.from_json(envelope)

    def sync(self) -> PNWhereNowResultEnvelope:
        return PNWhereNowResultEnvelope(super().sync())

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNWhereNowOperation

    def name(self):
        return "WhereNow"
